#!/usr/bin/perl

# $Id: $

########################################################################
# MAIA MAILGUARD LICENSE v.1.0
#
# Copyright 2004 by Robert LeBlanc <rjl@renaissoft.com>
#                   David Morton   <mortonda@dgrmm.net>
# All rights reserved.
#
# PREAMBLE
#
# This License is designed for users of Maia Mailguard
# ("the Software") who wish to support the Maia Mailguard project by
# leaving "Maia Mailguard" branding information in the HTML output
# of the pages generated by the Software, and providing links back
# to the Maia Mailguard home page.  Users who wish to remove this
# branding information should contact the copyright owner to obtain
# a Rebranding License.
#
# DEFINITION OF TERMS
#
# The "Software" refers to Maia Mailguard, including all of the
# associated PHP, Perl, and SQL scripts, documentation files, graphic
# icons and logo images.
#
# GRANT OF LICENSE
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. The end-user documentation included with the redistribution, if
#    any, must include the following acknowledgment:
#
#    "This product includes software developed by Robert LeBlanc
#    <rjl@renaissoft.com>."
#
#    Alternately, this acknowledgment may appear in the software itself,
#    if and wherever such third-party acknowledgments normally appear.
#
# 4. At least one of the following branding conventions must be used:
#
#    a. The Maia Mailguard logo appears in the page-top banner of
#       all HTML output pages in an unmodified form, and links
#       directly to the Maia Mailguard home page; or
#
#    b. The "Powered by Maia Mailguard" graphic appears in the HTML
#       output of all gateway pages that lead to this software,
#       linking directly to the Maia Mailguard home page; or
#
#    c. A separate Rebranding License is obtained from the copyright
#       owner, exempting the Licensee from 4(a) and 4(b), subject to
#       the additional conditions laid out in that license document.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
########################################################################

use Getopt::Long;
use DBI;
use Net::SMTP;
use Digest::MD5;


# prototypes
sub add_address($$$$$$$);
sub add_address_to_user($$$$$$$);
sub add_domain($$$$);
sub add_policy($$$$);
sub address_domain($);
sub clear_awl($$$);
sub clear_bayes($$$);
sub dedupe($);
sub delete_address($$$$);
sub delete_domain($$$$);
sub delete_domain_admin_references($$$$);
sub delete_mail($$$$);
sub delete_mail_reference($$$$);
sub delete_user_mail_references($$$$);
sub delete_user_wb_entries($$$$);
sub delete_wb_entry($$$$$);
sub expire_bayes($$$);
sub fatal($);
sub filter_addresses($$);
sub filter_domains($$);
sub generate_random_password();
sub get_address_type($);
sub get_email_address_priority($);
sub get_maia_user_row($$$$);
sub get_rewritten_address($$$);
sub get_user_from_address($$);
sub get_user_id($$$$$);
sub issue_password($$$);
sub link_address($$$$$);
sub output($);
sub prune_awl($$$$);
sub read_input_file($$$);
sub read_link_file($$);
sub send_email($$$$$);
sub unconflict($$$);
sub valid_address($);
sub valid_domain($);


# name of this script
my $script_name = "maiadbtool";

# read configuration file (/etc/maia.conf)
my $config_file = "/etc/maia.conf";
unless (my $rv = do $config_file) {
    fatal(sprintf("Couldn't parse %s: %s", $config_file, $@)) if $@;
    fatal(sprintf("Couldn't open %s", $config_file)) if (!defined($rv) || !$rv);
};

# defaults (overridden by settings in /etc/maia.conf)
$sa_learn = "sa-learn" if !defined($sa_learn);
$address_rewriting_type = 0 if !defined($address_rewriting_type);
$routing_domain = "" if !defined($routing_domain);
$auth_method = "internal" if !defined($auth_method);


my @add_addresses = ();
my @delete_addresses = ();
my @add_domains = ();
my @delete_domains = ();
my @link_addresses = ();
my $link_account = undef;
my $add_from_file = undef;
my $delete_from_file = undef;
my $link_from_file = undef;
my $clear_bayes = 0;
my $clear_awl = 0;
my $expire_bayes = 0;
my $prune_awl = undef;
my $help = 0;
my $debug = 0;
my $quiet = 0;

GetOptions("add-address|add-addresses=s" => \@add_addresses,
           "delete-address|del-address|delete-addresses|del-addresses=s" => \@delete_addresses,
           "link-address|link-addresses=s" => \@link_addresses,
           "link-account=s" => \$link_account,
           "add-domain|add-domains=s" => \@add_domains,
           "delete-domain|del-domain|delete-domains|del-domains=s" => \@delete_domains,
           "add-from-file=s" => \$add_from_file,
           "delete-from-file|del-from-file=s" => \$delete_from_file,
           "link-from-file=s" => \$link_from_file,
           "address-rewriting-type=i" => \$address_rewriting_type,
           "routing-domain=s" => \$routing_domain,
           "auth-method|authentication-method=s" => \$auth_method,
           "clear-bayes" => \$clear_bayes,
           "clear-awl" => \$clear_awl,
           "expire-bayes" => \$expire_bayes,
           "prune-awl:s" => \$prune_awl,
           "sa-learn=s" => \$sa_learn,
           "help" => \$help,
           "debug" => \$debug,
           "quiet" => \$quiet);

@add_addresses = split(/,/, join(',', @add_addresses));
@delete_addresses = split(/,/, join(',', @delete_addresses));
@add_domains = split(/,/, join(',', @add_domains));
@delete_domains = split(/,/, join(',', @delete_domains));
@link_addresses = split(/,/, join(',', @link_addresses));

# Display usage information
if ($help) {
    output("maiadbtool.pl\n" .
           "   --add-address addr1,addr2,...      : add one or more user e-mail addresses\n" .
           "   --add-addresses addr1,addr2,...    : same as --add-address\n" .
           "   --delete-address addr1,addr2,...   : delete one or more user e-mail addresses\n" .
           "   --del-address addr1,addr2,...      : same as --delete-address\n" .
           "   --delete-addresses addr1,addr2,... : same as --delete-address\n" .
           "   --del-addresses addr1,addr2,...    : same as --delete-address\n" .
           "   --link-address addr1,addr2,...     : link one or more addresses to a user account\n" .
           "   --link-addresses addr1,addr2,...   : same as --link-address\n" .
           "   --link-account address             : address of user to whom addresses should be linked\n" .
           "   --add-domain dom1,dom2,...         : add one or more domains\n" .
           "   --add-domains dom1,dom2,...        : same as --add-domain\n" .
           "   --delete-domain dom1,dom2,...      : delete one or more domains and all e-mail addresses in those domains\n" .
           "   --del-domain dom1,dom2,...         : same as --delete-domain\n" .
           "   --delete-domains dom1,dom2,...     : same as --delete-domain\n" .
           "   --del-domains dom1,dom2,...        : same as --delete-domain\n" .
           "   --add-from-file file               : add users and/or domains from a file\n" .
           "   --delete-from-file file            : delete users and/or domains from a file\n" .
           "   --del-from-file file               : same as --delete-from-file\n" .
           "   --link-from-file file              : link addresses to user accounts from a file\n" .
           "   --address-rewriting-type code      : address rewriting method [0..5] (see config.php)\n" .
           "   --routing-domain domain            : routing domain (see config.php)\n" .
           "   --auth-method method               : authentication method (see config.php)\n" .
           "   --authentication-method method     : same as --auth-method\n" .
           "   --clear-bayes                      : empty the Bayes database\n" .
           "   --clear-awl                        : empty the AWL database\n" .
           "   --expire-bayes                     : expire Bayes tokens\n" .
           "   --prune-awl [min]                  : prune the AWL database, removing entries with less than [min] occurrences\n" .
           "   --sa-learn path                    : path to your sa-learn script\n" .
           "   --help                             : display this help text\n" .
           "   --debug                            : display detailed debugging information\n" .
           "   --quiet                            : display only error messages");
    exit;
}

if (!defined($prune_awl)) {
    $prune_awl = 0;
} elsif ($prune_awl eq "") {
    $prune_awl = 2; # default minimum
}

# Resolve any debug/quiet conflicts
if ($debug && $quiet) {
    $debug = 0;
    $quiet = 0;
    output("Warning: --debug and --quiet negate each other.");
}

if ($clear_bayes && $expire_bayes) {
    $expire_bayes = 0;
    output("Warning: no point expiring the Bayes database when it's being cleared as well.");
}

if ($clear_awl && $prune_awl) {
    $prune_awl = 0;
    output("Warning: no point pruning the AWL database when it's being cleared as well.");
}

if (@link_addresses && !defined($link_account)) {
    @link_addresses = ();
    fatal("--link-account must be specified with --link-address.");
}

# database configuration
if (defined($dsn) && defined($username) && defined($password)) {
    $dbh = DBI->connect($dsn, $username, $password)
        or fatal("Can't connect to the Maia database (verify \$dsn, \$username, and \$password in maia.conf)");
} else {
    fatal("Can't connect to the Maia database (missing \$dsn, \$username, or \$password in maia.conf)");
}


# Read the (optional) input files for addresses and domains to add or delete
read_input_file(\@add_addresses, \@add_domains, $add_from_file)
    if (defined($add_from_file));
read_input_file(\@delete_addresses, \@delete_domains, $delete_from_file)
    if (defined($delete_from_file));
read_link_file(\@link_addresses, $link_from_file)
    if (defined($link_from_file));

# Filter out any invalid addresses and domains
filter_addresses(\@add_addresses, $quiet);
filter_addresses(\@delete_addresses, $quiet);
filter_addresses(\@link_addresses, $quiet);
filter_domains(\@add_domains, $quiet);
filter_domains(\@delete_domains, $quiet);

# Remove any duplicates
dedupe(\@add_addresses);
dedupe(\@delete_addresses);
dedupe(\@link_addresses);
dedupe(\@add_domains);
dedupe(\@delete_domains);

# Identify any conflicting addresses and domains
unconflict(\@add_addresses, \@delete_addresses, $debug);
unconflict(\@add_domains, \@delete_domains, $debug);

# Delete domains
foreach my $domain (@delete_domains) {
    delete_domain($dbh, $domain, $debug, $quiet);
}

# Add domains
foreach my $domain (@add_domains) {
    add_domain($dbh, $domain, $debug, $quiet);
}

# Delete addresses
foreach my $address (@delete_addresses) {
    delete_address($dbh, $address, $debug, $quiet);
}

# Add addresses
foreach my $address (@add_addresses) {
    add_address($dbh, $address, $address_rewriting_type, 
                $routing_domain, $auth_method, $debug, $quiet);
}

# Link addresses
foreach my $address (@link_addresses) {
    link_address($dbh, $address, $account, $debug, $quiet);
}

# Bayes clear/expire
if ($clear_bayes) {
    clear_bayes($dbh, $debug, $quiet);
} elsif ($expire_bayes) {
    expire_bayes($sa_learn, $debug, $quiet);
}

# AWL clear/prune
if ($clear_awl) {
    clear_awl($dbh, $debug, $quiet);
} elsif ($prune_awl) {
    prune_awl($prune_awl, $dbh, $debug, $quiet);
}

# Disconnect from the database
$dbh->disconnect;

exit;


########################################################################
###              End of Script: Subroutines begin below              ###
########################################################################

# Die, printing a time-stamped error message.
sub fatal($) {
    my ($msg) = @_;

    output("FATAL ERROR: " . $msg);
    exit 1;
}


# Write a time-stamped string to stdout for logging purposes.
sub output($) {
    my ($msg) = @_;
    my ($year, $month, $day, $hour, $minute, $second);
    my ($second, $minute, $hour, $day, $month, $year) = (localtime)[0,1,2,3,4,5];

    printf("%04d-%02d-%02d %02d:%02d:%02d Maia: [%s] %s\n",
           $year+1900, $month+1, $day, $hour, $minute, $second, $script_name, $msg);
}


# Verify that an e-mail address is syntactically valid
sub valid_address($) {
    my($address) = @_;
   
    return ($address =~ /^([0-9a-zA-Z]([-.\w]*[0-9a-zA-Z])*@([0-9a-zA-Z][-\w]*[0-9a-zA-Z]\.)+[a-zA-Z]{2,9})$/);
}


# Verify that a domain name is syntactically valid
sub valid_domain($) {
    my($domain) = @_;

    return ($domain =~ /^@([0-9a-zA-Z][-\w]*[0-9a-zA-Z]\.)+[a-zA-Z]{2,9}$/);
}


# Remove any invalid e-mail addresses from the list
sub filter_addresses($$) {
    my($address_ref, $quiet) = @_;

    my @valid = ();
    foreach my $address (@$address_ref) {
        if (valid_address($address)) {
            push(@valid, $address);
        } else {
            output("Warning: Skipping invalid address '$address'")
                if (!$quiet);
        }
    }
    @$address_ref = @valid;
}


# Remove any invalid domain names from the list
sub filter_domains($$) {
    my($domain_ref, $quiet) = @_;

    my @valid = ();
    foreach my $domain (@$domain_ref) {
        if (!($domain =~ /^@/)) {
            $domain = "@" . $domain;
        }
        if (valid_domain($domain)) {
            push(@valid, $domain);
        } else {
            output("Warning: Skipping invalid domain '$domain'")
                if (!$quiet);
        }
    }
    @$domain_ref = @valid;
}


# Remove any duplicate array elements
sub dedupe($) {
    my($array_ref) = @_;

    my %seen = ();
    foreach my $element (@$array_ref) {
        $seen{$element}++;
    }
    @$array_ref = keys %seen;
}


# Remove any array elements found on both the addition and deletion lists,
# since these would cancel each other out.
sub unconflict($$$) {
    my($add_ref, $del_ref, $debug) = @_;

    my %adding = ();
    my %deleting = ();
    foreach my $add (@$add_ref) {
        $adding{$add} = 1;
    }
    foreach my $del (@$del_ref) {
        $deleting{$del} = 1;
    }

    my @add_no_conflicts = ();
    foreach my $add (@$add_ref) {
        if (!$deleting{$add}) {
            push(@add_no_conflicts, $add);
        } else {
            output("Ignoring '$add': adding and deleting negate one another")
                if ($debug);
        }
    }
    @$add_ref = @add_no_conflicts;

    my @delete_no_conflicts = ();
    foreach my $del (@$del_ref) {
        if (!$adding{$del}) {
            push(@delete_no_conflicts, $del);
        }
    }
    @$del_ref = @delete_no_conflicts;
}


# Parse an input file for e-mail addresses and domain names, supplying a
# leading '@' for domain names that lack it.  Blank lines and whitespace
# should be ignored.
sub read_input_file($$$) {
    my($addresses_ref, $domains_ref, $infile) = @_;

    open (INFILE, "< $infile")
        or fatal("Can't read input file $infile: $!");
    while (<INFILE>) {
        chomp;
        if ($_ =~ /^\s*(.+@.+)\s*$/) {
            push(@$addresses_ref, $1);
        } elsif ($_ =~ /^\s*(@.+)\s*$/) {
            push(@$domains_ref, $1);
        } elsif ($_ =~ /^\s*([A-Za-z0-9\.\-]+)\s*$/) {
            push(@$domains_ref, "@" . $1);
        } else {
            # skip blank lines
        }
    }
    close INFILE;
}


# Parse an input file for lines containing a username or e-mail address
# followed by one or more e-mail addresses to link to it, separated by
# whitespace.  Blank lines and leading and trailing whitespace should be
# ignored.
sub read_link_file($$) {
    my($addresses_ref, $infile) = @_;

    open (INFILE, "< $infile")
        or fatal("Can't read input file $infile: $!");
    while (<INFILE>) {
        chomp;
        

    }
    close INFILE;
}


# Delete a domain, and any e-mail addresses in that domain
sub delete_domain($$$$) {
    my($dbh, $domain, $debug, $quiet) = @_;
    my($sth, $query, $address);

    output("Deleting domain '$domain'") if (!$quiet);

    # delete any addresses in the domain
    $query = "SELECT email FROM users WHERE email LIKE '_%$domain'";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    while (($address) = $sth->fetchrow()) {
        delete_address($dbh, $address, $debug, $quiet);
    }
    $sth->finish;

    # delete any references to this domain in the maia_domain_admins table
    delete_domain_admin_references($dbh, $domain, $debug, $quiet);

    output("Deleting domain record from maia_domains table") if ($debug);
    $query = "DELETE FROM maia_domains WHERE domain = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Deleting domain record from maia_users table") if ($debug);
    $query = "DELETE FROM maia_users WHERE user_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Deleting domain record from users table") if ($debug);
    $query = "DELETE FROM users WHERE email = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Deleting domain record from policy table") if ($debug);
    $query = "DELETE FROM policy WHERE policy_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

}


# Delete any domain administrator records associated with a domain
sub delete_domain_admin_references($$$$) {
    my($dbh, $domain, $debug, $quiet) = @_;
    my($sth, $sth2, $sth3, $query, $admin_id, $domain_id, $other_domain_id);

    # list all the admins that reference the target domain
    $query = "SELECT maia_domain_admins.admin_id, maia_domain_admins.domain_id " .
             "FROM maia_domain_admins, maia_domains " .
             "WHERE maia_domain_admins.domain_id = maia_domains.id " .
             "AND maia_domains.domain = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    while (($admin_id, $domain_id) = $sth->fetchrow()) {

        # delete this admin's reference to the target domain
        $query = "DELETE FROM maia_domain_admins WHERE domain_id = ? AND admin_id = ?";
        $sth2 = $dbh->prepare($query)
                   or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($domain_id, $admin_id)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        # does this admin administer any other domains?
        $query = "SELECT domain_id FROM maia_domain_admins WHERE admin_id = ?";
        $sth2 = $dbh->prepare($query)
                   or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($admin_id)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        if (!(($other_domain_id) = $sth2->fetchrow())) {

            # no - return his privilege level to that of a normal user
            $query = "UPDATE maia_users SET user_level = 'U' WHERE id = ?";
            $sth3 = $dbh->prepare($query)
                       or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
            $sth3->execute($admin_id)
                or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        }
        $sth2->finish;


    }
    $sth->finish;
}


# Create a new policy record for an address or domain, based on defaults
# inherited from the domain-default or system-default policies.
sub add_policy($$$$) {
    my($dbh, $address, $debug, $quiet) = @_;
    my($sth, $query, $domain, $policy_id, $nodefault);
    my($virus_lover, $spam_lover, $banned_files_lover, $bad_header_lover);
    my($bypass_virus_checks, $bypass_spam_checks, $bypass_banned_checks, $bypass_header_checks);
    my($discard_viruses, $discard_spam, $discard_banned_files, $discard_bad_headers);
    my($spam_modifies_subj, $spam_tag_level, $spam_tag2_level, $spam_kill_level);

    $policy_id = 0;
    $domain = address_domain($address);
    $query = "SELECT virus_lover, " .
                    "spam_lover, " .
                    "banned_files_lover, " .
                    "bad_header_lover, " .
                    "bypass_virus_checks, " .
                    "bypass_spam_checks, " .
                    "bypass_banned_checks, " .
                    "bypass_header_checks, " .
                    "discard_viruses, " .
                    "discard_spam, " .
                    "discard_banned_files, " .
                    "discard_bad_headers, " .
                    "spam_modifies_subj, " .
                    "spam_tag_level, " .
                    "spam_tag2_level, " .
                    "spam_kill_level " .
             "FROM policy WHERE policy_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));

    # try to find a domain-based set of defaults for this user, based on his e-mail address
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    if (($virus_lover, $spam_lover, $banned_files_lover, $bad_header_lover,
         $bypass_virus_checks, $bypass_spam_checks, $bypass_banned_checks, $bypass_header_checks,
         $discard_viruses, $discard_spam, $discard_banned_files, $discard_bad_headers,
         $spam_modifies_subj, $spam_tag_level, $spam_tag2_level, $spam_kill_level) = $sth->fetchrow()) {

        $nodefault = 0;

    } else {

        # try to find a "Default" policy (@.) to copy defaults from
        $sth->execute('Default')
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
        
        if (($virus_lover, $spam_lover, $banned_files_lover, $bad_header_lover,
             $bypass_virus_checks, $bypass_spam_checks, $bypass_banned_checks, $bypass_header_checks,
             $discard_viruses, $discard_spam, $discard_banned_files, $discard_bad_headers,
             $spam_modifies_subj, $spam_tag_level, $spam_tag2_level, $spam_kill_level) = $sth->fetchrow()) {

            $nodefault = 0;

        } else {

            # no suitable defaults found (should never get here)
            $nodefault = 1;
        }
    }
    $sth->finish;

    if ($nodefault) {

        # use the database defaults as our last resort.
        $query = "INSERT INTO policy (policy_name) VALUES (?)";
        $sth = $dbh->prepare($query)
            or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth->execute($address)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    } else {

        # use the domain or system default values found above.
        $query = "INSERT INTO policy (policy_name, " .
                                     "virus_lover, " .
                                     "spam_lover, " .
                                     "banned_files_lover, " .
                                     "bad_header_lover, " .
                                     "bypass_virus_checks, " .
                                     "bypass_spam_checks, " .
                                     "bypass_banned_checks, " .
                                     "bypass_header_checks, " .
                                     "discard_viruses, " .
                                     "discard_spam, " .
                                     "discard_banned_files, " .
                                     "discard_bad_headers, " .
                                     "spam_modifies_subj, " .
                                     "spam_tag_level, " .
                                     "spam_tag2_level, " .
                                     "spam_kill_level" .
                 ") VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
        $sth = $dbh->prepare($query)
            or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth->execute($address,
                      $virus_lover,
                      $spam_lover,
                      $banned_files_lover,
                      $bad_header_lover,
                      $bypass_virus_checks,
                      $bypass_spam_checks,
                      $bypass_banned_checks,
                      $bypass_header_checks,
                      $discard_viruses,
                      $discard_spam,
                      $discard_banned_files,
                      $discard_bad_headers,
                      $spam_modifies_subj,
                      $spam_tag_level,
                      $spam_tag2_level,
                      $spam_kill_level)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    }

    $query = "SELECT id FROM policy WHERE policy_name = ?";
    $sth = $dbh->prepare($query)
        or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($address)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($policy_id) = $sth->fetchrow())) {
        fatal("Couldn't find address '$address' in the policy table after insertion!");
    }
    $sth->finish;

    return $policy_id;
}


# Determine the address type:
#
#   0 : a plain domain name (e.g. example.com)
#   1 : a domain name with leading '@' (e.g. @example.com)
#   2 : an email address (e.g. joe@example.com)
sub get_address_type($) {
    my($address) = @_;

    my $pos = index($address, '@');
    if ($pos == -1) {      # "domain"
        return 0;
    } elsif ($pos > 0) {   # "user@domain"
        return 2;
    } else {               # "@domain"
        return 1;
    }
}


# Determine the priority value of an email address, based on
# how 'specific' it is.  The more dots it contains, the more
# specific we decide it to be.  Full email addresses (as
# opposed to domain names) are given additional priority.
sub get_email_address_priority($) {
    my($address) = @_;
    my($base);

    # give user@domain addresses more priority than @domain types
    my $address_type = get_address_type($address);
    if ($address_type == 2) {   # "user@domain"
        $base = 10;
    } else {
        $base = 0;
    }

    # give more specific addresses (i.e. addresses with more '.' in them)
    # higher priority.
    my $dot_count = 0;
    while ($address =~ /\./g) {
        $dot_count++;
    }

    return ($base + 2 * $dot_count);
}


# Determine the username to use for authentication purposes from
# an email address.  In most cases this is everything to the left of
# the last '@' in the address, but when $address_rewriting_type is 4
# the entire email address is treated as the username.
sub get_user_from_address($$) {
    my($address, $address_rewriting_type) = @_;

    if ($address_rewriting_type == 4) {  # entire address is the username
        return $address;
    } elsif ($address =~ /^(.+)@.+$/) {
        return $1;
    }
    return undef;
}


# Rewrite an email address according to the rules specified by the
# $address_rewriting_type:
#
#    0 : no rewriting is performed
#    1 : user@domain becomes user.domain@$routing_domain
#    2 : user@domain becomes user_domain@$routing_domain
#    3 : user@domain becomes user@domain@$routing_domain
#    4 : no rewriting is performed
#    5 : user@domain becomes user@$routing_domain
sub get_rewritten_address($$$) {
    my($address, $address_rewriting_type, $routing_domain) = @_;
    my($new_address);

    if ($address_rewriting_type == 1) {   # user.domain@isp

        $new_address = $address;
        $new_address =~ s/\@/\./g;
        $new_address .= "@" . $routing_domain;

    } elsif ($address_rewriting_type == 2) {   # user_domain@isp

        $new_address = $address;
        $new_address =~ s/[\@\.]/\_/g;
        $new_address .= "@" . $routing_domain;

    } elsif ($address_rewriting_type == 3) {   # user@domain@isp

        $new_address = $address . "@" . $routing_domain;

    } elsif ($address_rewriting_type == 5) {   # user@isp

        $new_address = get_user_from_address($address, $address_rewriting_type) . "@" . $routing_domain;

    } else {   # no rewriting for types 0 and 4

        $new_address = $address;

    }

    return $new_address;
}


# Adds an email address to the users table, linking it to a Maia user account
sub add_address_to_user($$$$$$$) {
    my($dbh, $policy_id, $address, $user_id, $domain_id, $debug, $quiet) = @_;

    output("Adding address '$address' to user account '$user_id'") if ($debug);
    $priority = get_email_address_priority($address);
    $query = "INSERT INTO users (policy_id, email, priority, maia_user_id, maia_domain_id) VALUES (?,?,?,?,?)";
    $sth = $dbh->prepare($query)
        or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($policy_id, $address, $priority, $user_id, $domain_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    $query = "SELECT id FROM users WHERE email = ?";
    $sth = $dbh->prepare($query)
        or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($address)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($address_id) = $sth->fetchrow())) {
        fatal("Couldn't find address '$address' in the users table after insertion!");
    }
    $sth->finish;

    return $address_id;
}


# Reads a single Maia user's configuration data into a single-element array
# containing a hash of all the column values from the maia_users table.
# Elements can be referenced as, e.g. @$row[0]->{'user_name'}
sub get_maia_user_row($$$$) {
    my($dbh, $user_id, $debug, $quiet) = @_;
       
    $query = "SELECT * FROM maia_users WHERE id = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($row) = $sth->fetchall_arrayref({}))) {
        fatal("Couldn't find user ID '$user_id' in the maia_users table!");
    }
    $sth->finish;

    return $row;
}


# Returns the ID of the Maia user account associated with a given
# username.
sub get_user_id($$$$$) {
    my($dbh, $user_name, $address, $debug, $quiet) = @_;

    my $user_id = 0;

    $query = "SELECT id FROM maia_users WHERE user_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_name)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # try to look the user up by name first
    if (!(($user_id) = $sth->fetchrow())) {

        # now try looking the user up by e-mail address,
        # in case the user was "auto-created".
        $sth->execute($address)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
 
        if (($user_id) = $sth->fetchrow()) {

            # set the user's proper name in the database
            # record, to speed up future lookups.
            $query = "UPDATE maia_users SET user_name = ? WHERE id = ?";
            $sth = $dbh->prepare($query)
                or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
            $sth->execute($user_name, $user_id)
                or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        }
    }

    return $user_id;
}


# Add a new domain
sub add_domain($$$$) {
    my($dbh, $domain, $debug, $quiet) = @_;
    my($sth, $query, $existing_domain_id, $default_user_autocreation_policy);
    my($domain_id, $policy_id, $primary_email_id, $default_user_id);
    my($default_user_config, $maia_user_id);

    # make sure the domain doesn't already exist
    output("Making sure the '$domain' doesn't already exist") if ($debug);
    $query = "SELECT id FROM maia_domains WHERE domain = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (($existing_domain_id) = $sth->fetchrow()) {
        output("Skipping domain '$domain', since it already exists") if (!$quiet);
        $sth->finish;
        return;
    } else {
        output("Adding domain '$domain'") if (!$quiet);
    }
    $sth->finish;

    # get global autocreation policy
    output("Looking up global user-autocreation policy") if ($debug);
    $query = "SELECT enable_user_autocreation FROM maia_domains WHERE domain = '@.'";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($default_user_autocreation_policy) = $sth->fetchrow())) {
        $default_user_autocreation_policy = 'N';
    }
    $sth->finish;

    # add the domain to the maia_domains table
    output("Adding '$domain' to the maia_domains table") if ($debug);
    $query = "INSERT INTO maia_domains (domain, enable_user_autocreation) VALUES (?,?)";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain, $default_user_autocreation_policy)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    $query = "SELECT id FROM maia_domains WHERE domain = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($domain_id) = $sth->fetchrow())) {
        fatal("Couldn't find domain '$domain' in the maia_domains table after insertion!");
    }

    # add a new policy for the domain, based on the system defaults
    $policy_id = add_policy($dbh, $domain, $debug, $quiet);

    # add the domain default address to the users table
    $primary_email_id = add_address_to_user($dbh, $policy_id, $domain, 0, $domain_id, $debug, $quiet);

    # add the domain default account to the maia_users table
    $default_user_id = get_user_id($dbh, "@.", "@.", $debug, $quiet);
    $default_user_config = get_maia_user_row($dbh, $default_user_id, $debug, $quiet);
    output("Adding a domain-default account to the maia_users table for '$domain'") if ($debug);
    $query = "INSERT INTO maia_users (user_name, primary_email_id, reminders, theme_id, " .
                                     "charts, language, charset, auto_whitelist, items_per_page, " .
                                     "quarantine_digest_interval, truncate_subject, truncate_email) " .
             "VALUES (?,?,'N',?,?,?,?,?,?,?,?,?)";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain,
                  $primary_email_id,
                  @$default_user_config[0]->{'theme_id'},
                  @$default_user_config[0]->{'charts'},
                  @$default_user_config[0]->{'language'},
                  @$default_user_config[0]->{'charset'},
                  @$default_user_config[0]->{'auto_whitelist'},
                  @$default_user_config[0]->{'items_per_page'},
                  @$default_user_config[0]->{'quarantine_digest_interval'},
                  @$default_user_config[0]->{'truncate_subject'},
                  @$default_user_config[0]->{'truncate_email'})
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    $query = "SELECT id FROM maia_users WHERE user_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($maia_user_id) = $sth->fetchrow())) {
        fatal("Couldn't find user '$domain' in the maia_users table after insertion!");
    }
    $sth->finish;

    # update the users table to link the e-mail address back to the domain
    output("Linking the '$domain' address to the '$domain' domain-default account") if ($debug);
    $query = "UPDATE users SET maia_user_id = ? WHERE id = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($maia_user_id, $primary_email_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

}


# Delete an e-mail address
sub delete_address($$$$) {
    my($dbh, $address, $debug, $quiet) = @_;
    my($sth, $sth2, $sth3, $query, $maia_user_id, $maia_user_name, $address_id, $new_primary_address);

    output("Deleting e-mail address '$address'") if (!$quiet);

    $query = "SELECT users.id, users.maia_user_id, maia_users.user_name, maia_users.primary_email_id " .
             "FROM users, maia_users " .
             "WHERE users.maia_user_id = maia_users.id " .
             "AND users.email = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($address)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (($address_id, $maia_user_id, $maia_user_name, $primary_email_id) = $sth->fetchrow()) {

        output("Deleting address record from policy table") if ($debug);
        $query = "DELETE FROM policy WHERE policy_name = ?";
        $sth2 = $dbh->prepare($query)
                   or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($address)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        output("Deleting address record from users table") if ($debug);
        $query = "DELETE FROM users WHERE email = ?";
        $sth2 = $dbh->prepare($query)
                   or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($address)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        # was this address the owner's primary address?
        if ($address_id == $primary_email_id) {

            # yes - does the address owner own any other addresses?
            $query = "SELECT id, email FROM users WHERE maia_user_id = ? ORDER BY priority DESC";
            $sth2 = $dbh->prepare($query)
                       or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
            $sth2->execute($maia_user_id)
                or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

            if (($other_address_id, $new_primary_address) = $sth2->fetchrow()) {

                # yes - make his highest-priority remaining address his new primary
                output("Making '$new_primary_address' the new primary address for user '$maia_user_name'") if ($debug);
                $query = "UPDATE maia_users SET primary_email_id = ? WHERE id = ?";
                $sth3 = $dbh->prepare($query)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
                $sth3->execute($other_address_id, $maia_user_id)
                    or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

                $sth2->finish;

            } else {
   
                # no - it's safe to delete the user account as well
                output("Deleting user account '$maia_user_name', since it owns no other addresses") if ($debug);
                $query = "DELETE FROM maia_users WHERE id = ?";
                $sth3 = $dbh->prepare($query)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
                $sth3->execute($maia_user_id)
                    or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

                output("Deleting administrator records for user account '$maia_user_name'") if ($debug);
                $query = "DELETE FROM maia_domain_admins WHERE admin_id = ?";
                $sth3 = $dbh->prepare($query)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
                $sth3->execute($maia_user_id)
                    or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

                output("Deleting current statistics for user account '$maia_user_name'") if ($debug);
                $query = "DELETE FROM maia_stats WHERE user_id = ?";
                $sth3 = $dbh->prepare($query)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
                $sth3->execute($maia_user_id)
                    or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

                output("Deleting statistics history for user account '$maia_user_name'") if ($debug);
                $query = "DELETE FROM maia_stats_history WHERE user_id = ?";
                $sth3 = $dbh->prepare($query)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
                $sth3->execute($maia_user_id)
                    or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

                delete_user_mail_references($dbh, $maia_user_id, $debug, $quiet);
                delete_user_wb_entries($dbh, $maia_user_id, $debug, $quiet);

            }
            
        }
        $sth->finish;

    } else {
        output("Address '$address' does not exist, skipping") if (!$quiet);
    }
}


# Returns a randomly-generated 8-character alphanumeric
# password and its 32-byte MD5 hash.
sub generate_random_password() {
    my(@chars, $password, $digest);
    my($md5) = Digest::MD5->new;

    @chars = ("A".."Z", "a".."z", 0..9);
    $password = join("", @chars[ map { rand @chars } (1..8) ]);
    $md5->add($password);
    $digest = $1 if $md5->hexdigest =~ /^([0-9a-f]{32})$/; # untaint

    return ($password, $digest);
}


# Sends an e-mail via the downstream MTA
sub send_email($$$$$) {
    my ($smtp_server, $smtp_port, $from, $to, $body) = @_;

    my($smtp) = Net::SMTP->new($smtp_server, Port => $smtp_port) ;
    fatal("Couldn't connect to SMTP server: $!") unless $smtp;
    $smtp->mail($from) ;
    $smtp->to($to) ;
    $smtp->data() ;
    $smtp->datasend($body . "\n") ;
    $smtp->dataend() ;
    $smtp->quit() ;
}


# Assigns a random password to a new user and sends
# the user an e-mail with temporary login credentials.
sub issue_password($$$) {
    my ($dbh, $user_email, $user_id) = @_;
    my ($select, $update, $sth, @row);
    my ($smtp_server, $smtp_port, $admin_email, $newuser_template_file, $login_url);
    $query = "SELECT admin_email, " .
                    "newuser_template_file, " .
                    "reminder_login_url, " .
                    "smtp_server, " .
                    "smtp_port " .
             "FROM maia_config WHERE id = 0";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (@row = $sth->fetchrow_array()) {
        $admin_email = $1 if $row[0] =~ /^(.+\@.+)$/si; # untaint
        $newuser_template_file = $1 if $row[1] =~ /^(.+)$/si; # untaint
        $login_url = $1 if $row[2] =~ /^(.+)$/si; # untaint
        $smtp_server = $1 if $row[3] =~ /^(.+)$/si; # untaint
        $smtp_port = $1 if $row[4] =~ /^([1-9]+[0-9]*)$/si; # untaint
    }
    $sth->finish;

    # Read the e-mail template file into memory once
    open TEMPLATEFILE, "<" . $newuser_template_file
        or fatal("Unable to open '$newuser_template_file': $!");
    my($template) = "";
    my($line);
    while ($line = <TEMPLATEFILE>) {
        $template .= $line;
    }
    close TEMPLATEFILE;

    my($password, $digest) = generate_random_password();

    $query = "UPDATE maia_users SET password = ? WHERE id = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($digest, $user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # Perform substitutions in the template
    my($body) = $template;
    $body =~ s/%%LOGIN%%/$user_email/g;
    $body =~ s/%%PASSWORD%%/$password/g;
    $body =~ s/%%ADMINEMAIL%%/$admin_email/g;
    $body =~ s/%%LOGINURL%%/$login_url/g;

    # Send the e-mail
    send_email($smtp_server, $smtp_port, $admin_email, $user_email, $body);
}


# Add an e-mail address, and its domain if necessary
sub add_address($$$$$$$) {
    my($dbh, $address, $address_rewriting_type, $routing_domain, $auth_method, $debug, $quiet) = @_;
    my($sth, $query, $domain, $default_domain_id, $default_domain_config);
    my($existing_address_id, $existing_domain_id, $user_name, $new_address);
    my($policy_id, $primary_address_id, $user_id);

    # determine the username and the rewritten e-mail address
    if ((lc($auth_method) eq 'pop3') && ($routing_domain ne '')) {
        $user_name = get_user_from_address($address, $address_rewriting_type);
        $new_address = $user_name . "@" . $routing_domain;
    } elsif (lc($auth_method) eq 'imap') {
        $new_address = get_rewritten_address($address, $address_rewriting_type, $routing_domain);
        $user_name = get_user_from_address($new_address, $address_rewriting_type);
    } elsif (lc($auth_method) eq 'internal') {
        $new_address = get_rewritten_address($address, $address_rewriting_type, $routing_domain);
        $user_name = $new_address;
    } else {
        $new_address = $address;
        $user_name = get_user_from_address($address, $address_rewriting_type);
    }
    if (!defined($user_name)) {
        output("Skipping invalid address '$address' (rewritten as '$new_address')") if (!$quiet);
        return;
    }
    if ($address ne $new_address) {
        output("Address '$address' rewritten as '$new_address'") if ($debug);
    }
    output("Username for authentication purposes is '$user_name'") if ($debug);

    # make sure the address doesn't already exist
    output("Making sure address '$new_address' doesn't already exist") if ($debug);
    $query = "SELECT id FROM users WHERE email = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($new_address)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (($existing_address_id) = $sth->fetchrow()) {
        output("Skipping address '$new_address', since it already exists") if (!$quiet);
        $sth->finish;
        return;
    } else {
        output("Adding address '$new_address'") if (!$quiet);
    }
    $sth->finish;

    # create the domain if necessary
    $domain = address_domain($new_address);
    output("Verifying that the '$domain' domain exists") if ($debug);
    $query = "SELECT id FROM maia_domains WHERE domain = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($domain)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($existing_domain_id) = $sth->fetchrow())) {
        add_domain($dbh, $domain, $debug, $quiet);
    }
    $sth->finish;

    # get domain-default configuration 
    $default_domain_id = get_user_id($dbh, $domain, $domain, $debug, $quiet);
    $default_domain_config = get_maia_user_row($dbh, $default_domain_id, $debug, $quiet);

    # add an entry to the maia_users table
    $query = "INSERT INTO maia_users (user_name, theme_id, reminders, charts, language, " .
                                     "charset, auto_whitelist, items_per_page, " .
                                     "quarantine_digest_interval, truncate_subject, truncate_email) " .
             "VALUES (?,?,?,?,?,?,?,?,?,?,?)";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_name,
                  @$default_domain_config[0]->{'theme_id'},
                  @$default_domain_config[0]->{'reminders'},
                  @$default_domain_config[0]->{'charts'},
                  @$default_domain_config[0]->{'language'},
                  @$default_domain_config[0]->{'charset'},
                  @$default_domain_config[0]->{'auto_whitelist'},
                  @$default_domain_config[0]->{'items_per_page'},
                  @$default_domain_config[0]->{'quarantine_digest_interval'},
                  @$default_domain_config[0]->{'truncate_subject'},
                  @$default_domain_config[0]->{'truncate_email'})
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # get the UID of this new record
    $query = "SELECT id FROM maia_users WHERE user_name = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_name)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($user_id) = $sth->fetchrow())) {
        fatal("Couldn't find user '$user_name' in the maia_users table after insertion!");
    }
    $sth->finish;

    # add a new policy for the address, based on the domain defaults
    $policy_id = add_policy($dbh, $new_address, $debug, $quiet);

    # add the address to the users table
    $primary_address_id = add_address_to_user($dbh, $policy_id, $new_address, $user_id, 0, $debug, $quiet);

    # Make this address the user's primary address
    $query = "UPDATE maia_users SET primary_email_id = ? WHERE id = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($primary_address_id, $user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # generate a random password and email it to the new user if necessary
    if (lc($auth_method) eq 'internal') {
        issue_password($dbh, $new_address, $user_id);
    }

    return $user_id;
}


# Link an e-mail address to a user account
sub link_address($$$$$) {
    my($dbh, $address, $account, $debug, $quiet) = @_;

   # output("Linking address '$address' to account '$account'")
   #     if (!$quiet);
   output("Address-linking has not been implemented yet.") if (!$quiet);

}


# Delete one or more cached/quarantined emails referenced by IDs
sub delete_mail($$$$) {
    my($dbh, $mail_ids_ref, $debug, $quiet) = @_;
    my($sth, $query, @mail_ids);

    @mail_ids = ();
    if (ref($mail_ids_ref) eq "ARRAY") {
        @mail_ids = @$mail_ids_ref;
    } else {
        push(@mail_ids, $mail_ids_ref);  # scalar implies a single mail ID
    }

    # delete any references to recipients
    $query = "DELETE FROM maia_mail_recipients " .
             "WHERE mail_id IN (?" . 
             ",?" x (scalar(@mail_ids) - 1) . 
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@mail_ids)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # delete any references to SpamAssassin rules
    $query = "DELETE FROM maia_sa_rules_triggered " .
             "WHERE mail_id IN (?" . 
             ",?" x (scalar(@mail_ids) - 1) . 
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@mail_ids)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # delete any references to viruses
    $query = "DELETE FROM maia_viruses_detected " .
             "WHERE mail_id IN (?" . 
             ",?" x (scalar(@mail_ids) - 1) .
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@mail_ids)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # delete any references to banned file attachments
    $query = "DELETE FROM maia_banned_attachments_found " .
             "WHERE mail_id IN (?" . 
             ",?" x (scalar(@mail_ids) - 1) .
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@mail_ids)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # delete the mail item itself
    $query = "DELETE FROM maia_mail " .
             "WHERE id IN (?" .
             ",?" x (scalar(@mail_ids) - 1) .
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@mail_ids)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

}


# Delete references to a specific set of mail items received by a 
# specific Maia user
sub delete_mail_references($$$$$) {
    my($dbh, $user_id, $mail_ids_ref, $debug, $quiet) = @_;
    my($sth, $query, @deletions, $maia_mail_id);

    $query = "DELETE FROM maia_mail_recipients " .
             "WHERE recipient_id = ? " .
             "AND mail_id IN (?" .
             ",?" x (scalar(@$mail_ids_ref) - 1) .
             ")";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_id, @$mail_ids_ref)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # delete any mail items without remaining recipient references
    $query = "SELECT maia_mail.id " .
             "FROM maia_mail, maia_mail_recipients " .
             "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
             "AND maia_mail.id IN (?" . 
             ",?" x (scalar(@$mail_ids_ref) - 1) .
             ") " .
             "AND maia_mail_recipients.mail_id IS NULL";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute(@$mail_ids_ref)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    @deletions = ();
    while (($maia_mail_id) = $sth->fetchrow()) {
        push(@deletions, $maia_mail_id);
    }
    $sth->finish;
    if (@deletions) {
        delete_mail($dbh, \@deletions, $debug, $quiet);
    }
}


# Delete references to any mail items received by a specific Maia user
sub delete_user_mail_references($$$$) {
    my($dbh, $user_id, $debug, $quiet) = @_;
    my($sth, $query, $mail_id, @mail_ids);

    $query = "SELECT mail_id FROM maia_mail_recipients WHERE recipient_id = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    @mail_ids = ();
    while (($mail_id) = $sth->fetchrow()) {
        push(@mail_ids, $mail_id);
    }
    $sth->finish;
    if (@mail_ids) {
        delete_mail_references($dbh, $user_id, \@mail_ids, $debug, $quiet);
    }
}


# Delete a single sender-recipient whitelist or blacklist entry
sub delete_wb_entry($$$$$) {
    my($dbh, $user_id, $addr_id, $debug, $quiet) = @_;
    my($sth, $sth2, $query, $wb);

    $query = "DELETE FROM wblist WHERE rid = ? AND sid = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_id, $addr_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # if there are no other references to this sender address,
    # remove it from the mailaddr table as well.
    $query = "SELECT wb FROM wblist WHERE sid = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($addr_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if (!(($wb) = $sth->fetchrow())) {
        $query = "DELETE FROM mailaddr WHERE id = ?";
        $sth2 = $dbh->prepare($query)
                   or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($addr_id)
            or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    }
    $sth->finish;
}


# Delete any whitelist/blacklist entries belonging to a specific Maia user
sub delete_user_wb_entries($$$$) {
    my($dbh, $user_id, $debug, $quiet) = @_;
    my($sth, $query, $sid);

    $query = "SELECT sid FROM wblist WHERE rid = ?";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    while (($sid) = $sth->fetchrow()) {
        delete_wb_entry($dbh, $user_id, $sid, $debug, $quiet);
    }
    $sth->finish;
}



# Returns the domain portion of an e-mail address, including the '@'
sub address_domain($) {
    my($address) = @_;

    if ($address =~ /^.*(@.+)$/) {
        return $1;
    }
    return undef;
}


# Empty the Bayes database of all token records
sub clear_bayes($$$) {
    my($dbh, $debug, $quiet) = @_;
    my($sth, $query);

    output("Erasing the SpamAssassin Bayes database")
        if (!$quiet);

    output("Clearing the maia.bayes_token table")
        if ($debug);
    $query = "DELETE FROM bayes_token";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Clearing the maia.bayes_seen table")
        if ($debug);
    $query = "DELETE FROM bayes_seen";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Clearing the maia.bayes_vars table")
        if ($debug);
    $query = "DELETE FROM bayes_vars";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    output("Clearing the maia.bayes_expire table")
        if ($debug);
    $query = "DELETE FROM bayes_expire";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

}


# Empty the AWL database of all sender address records
sub clear_awl($$$) {
    my($dbh, $debug, $quiet) = @_;
    my($sth, $query);

    output("Erasing the SpamAssassin Auto-Whitelist (AWL) database")
        if (!$quiet);

    output("Clearing the maia.awl table")
        if ($debug);
    $query = "DELETE FROM awl";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
}


# Expire obsolete and/or statistically insignificant tokens from the Bayes
# database.
sub expire_bayes($$$) {
    my($sa_learn, $debug, $quiet) = @_;

    output("Expiring tokens from the SpamAssassin Bayes database")
        if (!$quiet);
    output("Using $sa_learn")
        if ($debug);

    my $command = "$sa_learn " . ($debug ? "-D " : "") . "--sync --force-expire 2>&1";
    my $result = `$command`;
    if ($? == -1) {
    	fatal("Failed to execute $sa_learn: $!");
    }
    if (!$quiet) {
        my @lines = split(/\n/, $result);
        foreach $line (@lines) {
            output($line);
        }
    }
}


# Prune the AWL database by removing sender records that occur less than
# a certain number of times (default: 2), and are associated with positive
# scores.
sub prune_awl($$$$) {
    my($prune_awl, $dbh, $debug, $quiet) = @_;
    my($sth, $query);

    output("Pruning the SpamAssassin Auto-Whitelist (AWL) database")
        if (!$quiet);
    output("Positive-scoring senders that appear less than $prune_awl times in the AWL will be deleted")
        if ($debug);

    $query = "DELETE FROM awl WHERE count < ? AND totscore >= 0";
    $sth = $dbh->prepare($query)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($prune_awl)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
}


