#!/usr/bin/perl

# $Id: process-quarantine-sub.pl 994 2006-04-06 06:11:49Z rjl $

########################################################################
# MAIA MAILGUARD LICENSE v.1.0
#
# Copyright 2004 by Robert LeBlanc <rjl@renaissoft.com>
#                   David Morton   <mortonda@dgrmm.net>
# All rights reserved.
#
# PREAMBLE
#
# This License is designed for users of Maia Mailguard
# ("the Software") who wish to support the Maia Mailguard project by
# leaving "Maia Mailguard" branding information in the HTML output
# of the pages generated by the Software, and providing links back
# to the Maia Mailguard home page.  Users who wish to remove this
# branding information should contact the copyright owner to obtain
# a Rebranding License.
#
# DEFINITION OF TERMS
#
# The "Software" refers to Maia Mailguard, including all of the
# associated PHP, Perl, and SQL scripts, documentation files, graphic
# icons and logo images.
#
# GRANT OF LICENSE
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. The end-user documentation included with the redistribution, if
#    any, must include the following acknowledgment:
#
#    "This product includes software developed by Robert LeBlanc
#    <rjl@renaissoft.com>."
#
#    Alternately, this acknowledgment may appear in the software itself,
#    if and wherever such third-party acknowledgments normally appear.
#
# 4. At least one of the following branding conventions must be used:
#
#    a. The Maia Mailguard logo appears in the page-top banner of
#       all HTML output pages in an unmodified form, and links
#       directly to the Maia Mailguard home page; or
#
#    b. The "Powered by Maia Mailguard" graphic appears in the HTML
#       output of all gateway pages that lead to this software,
#       linking directly to the Maia Mailguard home page; or
#
#    c. A separate Rebranding License is obtained from the copyright
#       owner, exempting the Licensee from 4(a) and 4(b), subject to
#       the additional conditions laid out in that license document.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
########################################################################

use DBI;
use Mail::SpamAssassin;
use Getopt::Long;
use Time::HiRes ();

# constants
use constant RAZOR => 1;
use constant PYZOR => 2;
use constant DCC => 4;
use constant SPAMCOP => 8;
use constant ALL => (RAZOR + PYZOR + DCC + SPAMCOP);

my $sa = Mail::SpamAssassin->new();
my $sa_version = $sa->VERSION;
if ($sa_version >= 3.000000) {
    require Mail::SpamAssassin::Message;
    Mail::SpamAssassin::Message->import();
} else {
    require Mail::SpamAssassin::NoMailAudit;
    Mail::SpamAssassin::NoMailAudit->import();
}
use Mail::SpamAssassin::PerMsgLearner;
use Mail::SpamAssassin::Reporter;

# prototypes
sub fatal($);
sub output($);
sub get_encryption_key($);
sub text_is_encrypted($);
sub decrypt_text($$$);
sub delete_mail_item($$$);
sub delete_mail($$$$);
sub do_learn($$$$);
sub do_report($$$$$);
sub process_spam($$$$$$$$$);
sub process_ham($$$$$$$);
sub init();
sub section_time($);
sub report();

# name of this script
my $script_name = "process-quarantine-sub";

# read configuration file (/etc/maia.conf)
my $config_file = "/etc/maia.conf";
unless (my $rv = do $config_file) {
    fatal(sprintf("Couldn't parse %s: %s", $config_file, $@)) if $@;
    fatal(sprintf("Couldn't open %s", $config_file)) if (!defined($rv) || !$rv);
};

# defaults (overridden by settings in /etc/maia.conf)
$default_limit = 5 if !defined($default_limit);
$key_file = undef if !defined($key_file);
$default_max_size = 256*1024 if !defined($default_max_size);
$learning_options = 0 if !defined($learning_options);
$report_options = 0 if !defined($report_options);

my $limit = $default_limit;
my $max_size = $default_max_size;
my $learn = $learning_options;
my $report = ($report_options > 0);
my $no_razor = !($report_options & RAZOR);
my $no_pyzor = !($report_options & PYZOR);
my $no_dcc = !($report_options & DCC);
my $no_spamcop = !($report_options & SPAMCOP);
my $spamonly = 0;  # disabled
my $hamonly = 0;   # disabled
my $debug = undef; # disabled
my $quiet = 0;     # disabled
@timing = ();

GetOptions("limit=i" => \$limit,         # --limit=int or --limit int
           "max-size=i" => \$max_size,   # --max-size=int or --max-size int
           "learn" => \$learn,           # --learn
           "report" => \$report,         # --report
           "no-razor" => \$no_razor,     # --no-razor
           "no-pyzor" => \$no_pyzor,     # --no-pyzor
           "no-dcc" => \$no_dcc,         # --no-dcc
           "no-spamcop" => \$no_spamcop, # --no-spamcop
           "spam-only" => \$spamonly,    # --spam-only
           "ham-only" => \$hamonly,      # --ham-only
           "debug:s" => \$debug,         # --debug, --debug=string or --debug string
           "quiet" => \$quiet);          # --quiet

# Sanity-check any supplied arguments
if (!defined($debug)) {
    $debug = 0;
} elsif ($debug eq "") {
    $debug = "info";
}
if ($debug && $quiet) {
    $debug = 0;
    $quiet = 0;
    output("Warning: --debug and --quiet negate each other.");
}
$limit = ($limit > 0 ? $limit : $default_limit);
if ($limit > 127) {
    $limit = 127;
    if ($debug) {
        output("Warning: enforcing maximum limit of 127 spam/non-spam items per run.");
    }
}
$max_size = ($max_size > 0 ? $max_size : $default_max_size);
if ($spamonly && $hamonly) {
    $spamonly = 0;
    $hamonly = 0;
    if ($debug) {
        output("Warning: --spam-only and --ham-only negate each other.");
    }
}
if ($hamonly) {
    $report = 0;
    if ($debug) {
        output("Warning: --report disregarded when --ham-only is specified.");
    }
}
if ($report && $no_razor && $no_pyzor && $no_dcc && $no_spamcop) {
    $report = 0;
    output("Warning: no one to report to, reporting is disabled.")
        if (!$quiet);
} elsif (!$report && ($no_razor || $no_pyzor || $no_dcc | $no_spamcop)) {
    $no_razor = 0;
    $no_pyzor = 0;
    $no_dcc = 0;
    $no_spamcop = 0;
    output("Warning: --report not specified, reporting is disabled.")
        if (!$quiet);
}
if ($debug eq "info" || $debug eq "all") {
    if ($sa_version < 3.001000) {
        $debug = 1;
        output("Warning: only one debugging mode available in SpamAssassin < 3.1.");
    }
} elsif ($debug) {
    output(sprintf("Warning: unknown debugging level '%s', using lowest setting.", $debug));
    $debug = (($sa_version >= 3.001000) ? "info" : 1);
}
if ($debug) {
    output("Starting");
    output(sprintf("limit = %d items per run", $limit));
    output(sprintf("max-size = %d bytes", $max_size));
    output(sprintf("learn = %s", ($learn ? "enabled" : "disabled")));
    output(sprintf("report = %s", ($report ? "enabled" : "disabled")));
    if ($report) {
        output(sprintf("report to Razor = %s", ($no_razor ? "disabled" : "enabled")));
        output(sprintf("report to Pyzor = %s", ($no_pyzor ? "disabled" : "enabled")));
        output(sprintf("report to DCC = %s", ($no_dcc ? "disabled" : "enabled")));
        output(sprintf("report to SpamCop = %s", ($no_spamcop ? "disabled" : "enabled")));
    }
    output(sprintf("spam-only = %s", ($spamonly ? "enabled" : "disabled")));
    output(sprintf("ham-only = %s", ($hamonly ? "enabled" : "disabled")));
    output(sprintf("debug = enabled (%s)", $debug));
}

my $report_targets = ALL - ($no_razor * RAZOR) 
                         - ($no_pyzor * PYZOR)
                         - ($no_dcc * DCC)
                         - ($no_spamcop * SPAMCOP);
my $dbh;

# database configuration
if (defined($dsn) && defined($username) && defined($password)) {
    $dbh = DBI->connect($dsn, $username, $password)
        or fatal("Can't connect to the Maia database (verify \$dsn, \$username, and \$password in maia.conf)");
} else {
    fatal("Can't connect to the Maia database (missing \$dsn, \$username, or \$password in maia.conf)");
}

# Process one batch of confirmed spam and/or confirmed ham
my $key = get_encryption_key($key_file);
my $spamcount = 0;
my $hamcount = 0;
if ($spamonly) {
    $spamcount = process_spam($dbh, $learn, $report, $key, $limit, $debug, $quiet, $max_size, $report_targets);
} elsif ($hamonly) {
    $hamcount = process_ham($dbh, $learn, $key, $limit, $debug, $quiet, $max_size);
} else {
    $spamcount = process_spam($dbh, $learn, $report, $key, $limit, $debug, $quiet, $max_size, $report_targets);
    $hamcount = process_ham($dbh, $learn, $key, $limit, $debug, $quiet, $max_size);
}

# Disconnect from the database
$dbh->disconnect;

# We're done, return the total number of items processed.
exit ($spamcount + $hamcount);


# Die, printing a time-stamped error message.
sub fatal($) {
    my ($msg) = @_;

    output("FATAL ERROR: " . $msg);
    exit 255;
}


# Write a time-stamped string to stdout for logging purposes.
sub output($) {
    my ($msg) = @_;
    my ($year, $month, $day, $hour, $minute, $second);
    my ($second, $minute, $hour, $day, $month, $year) = (localtime)[0,1,2,3,4,5];

    printf("%04d-%02d-%02d %02d:%02d:%02d Maia: [%s] %s\n",
           $year+1900, $month+1, $day, $hour, $minute, $second, $script_name, $msg);
}


# Read the encryption key from a file
sub get_encryption_key($) {
    my ($key_file) = @_;
    my ($key);

    if (!defined $key_file) {
        return undef;
    }

    use IO::File;
    my $fh = new IO::File;

    $key_file = $1 if $key_file =~ /^(.+)$/si; # untaint

    # Key file exists, read key from file
    if ($fh->open("<" . $key_file)) {
        sysread($fh, $key, 56);
        $fh->close;
        return $key;
    } else {
        fatal(sprintf("Can't open encryption key file %s", $key_file));
    }
}


# Returns true (1) if the text is encrypted, false (0) otherwise.
sub text_is_encrypted($) {
    my ($text) = @_;

    return ($text =~ /^RandomIV/);
}


# Decrypt a text string using a specified key.
sub decrypt_text($$$) {
    my ($key, $ciphertext, $debug) = @_;
    my ($cipher);

    require Crypt::CBC;

    $key = $1 if $key =~ /^([^\0]{56})$/si; # untaint
    if (text_is_encrypted($ciphertext)) {
        $cipher = Crypt::CBC->new( {'key' => $key,
                                    'cipher' => 'Blowfish',
                                    'regenerate_key' => 0,
                                    'padding' => 'null',
                                    'prepend_iv' => 1,
                                    'header' => 'randomiv'
                                 } );

        if ($debug) {
            output("Mail item is encrypted.");
        }
        return $cipher->decrypt($ciphertext);
    } else {
        if ($debug) {
            output("Mail item is not encrypted.");
        }
        return $ciphertext;
    }
}


# Delete one specific mail item and all references to it.
sub delete_mail_item($$$) {
    my($dbh, $mail_id, $debug) = @_;
    my($sth, $delete);

    $mail_id = $1 if $mail_id =~ /^([0-9]*)$/si; # untaint

    # Delete the mail item itself
    $delete = "DELETE FROM maia_mail WHERE id = ?";
    $sth = $dbh->prepare($delete)
                or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # Delete all recipient references to the mail item
    $delete = "DELETE FROM maia_mail_recipients WHERE mail_id = ?";
    $sth = $dbh->prepare($delete)
                or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # Delete any virus references to the mail item
    $delete = "DELETE FROM maia_viruses_detected WHERE mail_id = ?";
    $sth = $dbh->prepare($delete)
                or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # Delete any SpamAssassin rule references to the mail item
    $delete = "DELETE FROM maia_sa_rules_triggered WHERE mail_id = ?";
    $sth = $dbh->prepare($delete)
                 or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # Delete any banned file attachment references to the mail item
    $delete = "DELETE FROM maia_banned_attachments_found WHERE mail_id = ?";
    $sth = $dbh->prepare($delete)
                or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
}


# Delete mail references for recipients who agree that
# the mail item is [spam|ham].  Remove the mail item
# itself, if no other recipients reference it.
sub delete_mail($$$$) {
    my($dbh, $mail_id, $type, $debug) = @_;
    my($sth, $sth2, $delete, $select);

    $mail_id = $1 if $mail_id =~ /^([0-9]*)$/si; # untaint

    # Delete mail references for all recipients who agree
    # that the mail item is [spam|ham].
    $delete = "DELETE FROM maia_mail_recipients " .
              "WHERE mail_id = ? AND type = ?";
    $sth = $dbh->prepare($delete)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id, $type)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    if ($debug) {
        output(sprintf("Deleted spam/non-spam recipient references to mail item %d", $mail_id));
    }

    # See if any other mail references exist for this mail
    # item, i.e. any other recipients.
    $select = "SELECT recipient_id FROM maia_mail_recipients " .
              "WHERE mail_id = ?";
    $sth = $dbh->prepare($select)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($mail_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    # If no other recipients exist for this mail item,
    # delete the mail item itself.
    if (!$sth->fetchrow_array()) {
        delete_mail_item($dbh, $mail_id, $debug);
        if ($debug) {
            output(sprintf("Deleted mail item %d", $mail_id));
        }
    } else {
        if ($debug) {
            output(sprintf("Retaining mail item %d, since it still has recipient references.", $mail_id));
        }
    }
    $sth->finish;
}


# Train the Bayes engine explicitly with the contents of
# a single e-mail, forcing it to recognize the mail as
# spam or ham.
sub do_learn($$$$) {
    my($sa, $msg, $isspam, $debug) = @_;
    my($ma, $learner, $learned);

    my @body = split (/^/m, $msg);
    my $dataref = \@body;
    if ($sa_version >= 3.000000) {
        $ma = Mail::SpamAssassin::Message->new({message => $dataref, parse_now => 0});
    } else {
        $ma = Mail::SpamAssassin::NoMailAudit->new('data' => $dataref);
    }
    $ma->{noexit} = 1;
    section_time("msg-prep") if $debug;

    $learner = $sa->learn($ma, undef, $isspam, 0);
    $learned = $learner->did_learn();
    $learner->finish();
    $ma->finish();
    section_time("train-bayes") if $debug;

    return $learned;
}


# Report spam to Vipul's Razor, the DCC, and/or Pyzor,
# and optionally train the Bayes engine in the process
# (recognizing the mail item as spam).
sub do_report($$$$$) {
    my($sa, $msg, $learn, $debug, $report_targets) = @_;
    my($ma, $reported);
    my $no_razor = !($report_targets & RAZOR);
    my $no_pyzor = !($report_targets & PYZOR);
    my $no_dcc = !($report_targets & DCC);
    my $no_spamcop = !($report_targets & SPAMCOP);

    my @body = split (/^/m, $msg);
    my $dataref = \@body;
    if ($sa_version >= 3.000000) {
        $ma = Mail::SpamAssassin::Message->new({message => $dataref, parse_now => 0});
    } else {
        $ma = Mail::SpamAssassin::NoMailAudit->new ('data' => $dataref);
    }
    $ma->{noexit} = 1;
    $sa->{conf}->{bayes_learn_during_report} = $learn;
    section_time("msg-prep") if $debug;

    # SpamAssassin 3.1 and later reverse the return code
    # from report_as_spam().
    if ($sa_version >= 3.001000) {
        $reported = $sa->report_as_spam($ma, {
                                               dont_report_to_razor => $no_razor,
                                               dont_report_to_pyzor => $no_pyzor,
                                               dont_report_to_dcc => $no_dcc,
                                               dont_report_to_spamcop => $no_spamcop
                                             }
                                       );
    } else {
        $reported = !$sa->report_as_spam($ma, {
                                                dont_report_to_razor => $no_razor,
                                                dont_report_to_pyzor => $no_pyzor,
                                                dont_report_to_dcc => $no_dcc,
                                                dont_report_to_spamcop => $no_spamcop
                                              }
                                        );
    }
    $ma->finish();
    section_time("send-report") if $debug;

    return $reported;
}


# Process all the mail items in the database that
# recipients have confirmed as spam (type = 'C').
sub process_spam($$$$$$$$$) {
    my($dbh, $learn, $report, $key, $limit, $debug, $quiet, $max_size, $report_targets) = @_;
    my($sth, @row, $select);
    my($mail_id, $mail_contents, $mail_subject);

    # Assemble a list of all the confirmed spam (C) items
    $select = "SELECT maia_mail.id, " .
                     "maia_mail.contents, " .
                     "maia_mail.subject " .
              "FROM maia_mail, maia_mail_recipients " .
              "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
              "AND maia_mail_recipients.type = 'C' LIMIT " . $limit;
    $sth = $dbh->prepare($select)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    my $count = 0;
    my $learnedcount = 0;
    my $reportedcount = 0;

    # Initialize SpamAssassin
    if ($sa_version >= 3.001000) {
        $sa_debug_level = ($debug ? $debug : "");
    } else {
        $sa_debug_level = $debug;
    }
    my $sa = new Mail::SpamAssassin({ debug => $sa_debug_level });
    $sa->init (1);

    while (@row = $sth->fetchrow_array()) {
        $mail_id = $row[0];
        $mail_contents = $row[1];
        $mail_subject = $row[2];

        timer_init() if $debug;
        if (length($mail_contents) > $max_size) {
            output(sprintf("Skipping mail item %d (too big: %d > %d)",
                   $mail_id, length($mail_contents), $max_size))
                if (!$quiet);
        } else {
            if (defined $key) {
                $mail_contents = decrypt_text($key, $mail_contents, $debug);
                section_time("decrypt-msg") if $debug;
            }
            if ($learn && !$report) {
                if (do_learn($sa, $mail_contents, 1, $debug)) {
                    $learnedcount++;
                    output(sprintf("Learned mail item %d as spam", $mail_id))
                        if (!$quiet);
                }
            } elsif ($report && !$learn) {
                if (do_report($sa, $mail_contents, 0, $debug, $report_targets)) {
                    $reportedcount++;
                    output(sprintf("Reported mail item %d", $mail_id))
                        if (!$quiet);
                }
            } elsif ($report && $learn) {
                if (do_report($sa, $mail_contents, 1, $debug, $report_targets)) {
                    $reportedcount++;
                    $learnedcount++;
                    output(sprintf("Learned mail item %d as spam and reported it", $mail_id))
                      if (!$quiet);
                }
            }
        }

        # Remove all recipient references to the mail item
        # (and the item itself, if no other recipients remain)
        delete_mail($dbh, $mail_id, "C", $debug);
        section_time("delete-mail") if $debug;

        $count++;

        if ($debug) {
            output(timer_report());
        }

    }
    $sth->finish;

    if ($learn) {

        # Resynchronize the Bayes engine after learning
        $sa->rebuild_learner_caches();
        $sa->finish_learner();
    }

    output(sprintf("%d spam %s processed (%d learned, %d reported)",
           $count, ($count == 1 ? "item" : "items"),
           $learnedcount, $reportedcount))
        if (!$quiet);


    return $count;
}


# Process all the mail items in the database that
# recipients have confirmed as ham (type = 'G').
sub process_ham($$$$$$$) {
    my($dbh, $learn, $key, $limit, $debug, $quiet, $max_size) = @_;
    my($sth, @row, $select);
    my($mail_id, $mail_contents, $mail_subject);

    # Assemble a list of all the confirmed ham (G) items
    $select = "SELECT maia_mail.id, " .
                     "maia_mail.contents, " .
                     "maia_mail.subject " .
              "FROM maia_mail, maia_mail_recipients " .
              "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
              "AND maia_mail_recipients.type = 'G' LIMIT " . $limit;
    $sth = $dbh->prepare($select)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

    my $count = 0;
    my $learnedcount = 0;

    # Initialize SpamAssassin
    if ($sa_version >= 3.001000) {
        $sa_debug_level = ($debug ? $debug : "");
    } else {
        $sa_debug_level = $debug;
    }
    my $sa = new Mail::SpamAssassin({debug => $sa_debug_level});
    $sa->init (1);

    while (@row = $sth->fetchrow_array()) {
        $mail_id = $row[0];
        $mail_contents = $row[1];
        $mail_subject = $row[2];

        timer_init() if $debug;
        if (length($mail_contents) > $max_size) {
            output(sprintf("Skipping mail item %d (too big: %d > %d)",
                   $mail_id, length($mail_contents), $max_size))
                if (!$quiet);
        } else {
            if ($learn) {
                if (defined $key) {
                    $mail_contents = decrypt_text($key, $mail_contents, $debug);
                    section_time("decrypt-msg") if $debug;
                }
                if (do_learn($sa, $mail_contents, 0, $debug)) {
                    $learnedcount++;
                    output(sprintf("Learned mail item %d as non-spam", $mail_id))
                        if (!$quiet);
                }
            }
        }

        # Remove all recipient references to the mail item
        # (and the item itself, if no other recipients remain)
        delete_mail($dbh, $mail_id, "G", $debug);
        section_time("delete-mail") if $debug;

        $count++;

        if ($debug) {
            output(timer_report());
        }

    }
    $sth->finish;

    if ($learn) {

        # Resynchronize the Bayes engine after learning
        $sa->rebuild_learner_caches();
        $sa->finish_learner();
    }

    output(sprintf("%d non-spam %s processed (%d learned)",
           $count, ($count == 1 ? "item" : "items"), $learnedcount))
        if (!$quiet);

    return $count;
}


# Restart the timer
sub timer_init() {
  @timing = (); section_time('init');
}


# Add the current time to the timer array with a text label
# to describe the section of the code traversed since the
# previous timestamp.
sub section_time($) {
  push(@timing, shift, Time::HiRes::time);
}


# Returns an amavisd-style timing report (code courtesy of
# Mark Martinec from amavisd-new).
sub timer_report() {
  section_time('rundown');
  my($notneeded, $t0) = (shift(@timing), shift(@timing));
  my($total) = $t0 <= 0 ? 0 : $timing[$#timing] - $t0;
  if ($total < 0.0000001) { $total = 0.0000001 }
  my(@sections);
  while (@timing) {
    my($section, $t) = (shift(@timing), shift(@timing));
    my($dt) = $t-$t0;
    $dt = 0  if $dt < 0;  # just in case (clock jumps)
    my($dtp) = $dt > $total ? 100 : $dt*100.0/$total;
    push(@sections, sprintf("%s: %.0f (%.0f%%)", $section, $dt*1000, $dtp));
    $t0 = $t;
  }
  sprintf("TIMING [total %.0f ms] - %s", $total * 1000, join(", ",@sections));
}
