#!/usr/bin/perl

# $Id: process-quarantine.pl 994 2006-04-06 06:11:49Z rjl $

########################################################################
# MAIA MAILGUARD LICENSE v.1.0
#
# Copyright 2004 by Robert LeBlanc <rjl@renaissoft.com>
#                   David Morton   <mortonda@dgrmm.net>
# All rights reserved.
#
# PREAMBLE
#
# This License is designed for users of Maia Mailguard
# ("the Software") who wish to support the Maia Mailguard project by
# leaving "Maia Mailguard" branding information in the HTML output
# of the pages generated by the Software, and providing links back
# to the Maia Mailguard home page.  Users who wish to remove this
# branding information should contact the copyright owner to obtain
# a Rebranding License.
#
# DEFINITION OF TERMS
#
# The "Software" refers to Maia Mailguard, including all of the
# associated PHP, Perl, and SQL scripts, documentation files, graphic
# icons and logo images.
#
# GRANT OF LICENSE
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. The end-user documentation included with the redistribution, if
#    any, must include the following acknowledgment:
#
#    "This product includes software developed by Robert LeBlanc
#    <rjl@renaissoft.com>."
#
#    Alternately, this acknowledgment may appear in the software itself,
#    if and wherever such third-party acknowledgments normally appear.
#
# 4. At least one of the following branding conventions must be used:
#
#    a. The Maia Mailguard logo appears in the page-top banner of
#       all HTML output pages in an unmodified form, and links
#       directly to the Maia Mailguard home page; or
#
#    b. The "Powered by Maia Mailguard" graphic appears in the HTML
#       output of all gateway pages that lead to this software,
#       linking directly to the Maia Mailguard home page; or
#
#    c. A separate Rebranding License is obtained from the copyright
#       owner, exempting the Licensee from 4(a) and 4(b), subject to
#       the additional conditions laid out in that license document.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
########################################################################

use Getopt::Long;
use DBI;

# prototypes
sub output($);
sub fatal($);
sub update_mail_stats($$$);
sub recalc_stats($);

# name of this script
my $script_name = "process-quarantine";

# read configuration file (/etc/maia.conf)
my $config_file = "/etc/maia.conf";
unless (my $rv = do $config_file) {
    fatal(sprintf("Couldn't parse %s: %s", $config_file, $@)) if $@;
    fatal(sprintf("Couldn't open %s", $config_file)) if (!defined($rv) || !$rv);
};

# defaults (overridden by settings in /etc/maia.conf)
$script_dir = "/var/amavisd/maia/scripts" if !defined($script_dir);
$pid_file = "/var/amavisd/.process-quarantine.pid" if !defined($pid_file);

my $limit = 0;
my $max_size = 0;
my $learn = 0;
my $report = 0;
my $no_razor = 0;
my $no_pyzor = 0;
my $no_dcc = 0;
my $no_spamcop = 0;
my $spamonly = 0;
my $hamonly = 0;
my $help = 0;
my $debug = undef;
my $quiet = 0;

GetOptions("limit=i" => \$limit,         # --limit=int or --limit int
           "max-size=i" => \$max_size,   # --max-size=int or --max-size int
           "learn" => \$learn,           # --learn
           "report" => \$report,         # --report
           "no-razor" => \$no_razor,     # --no-razor
           "no-pyzor" => \$no_pyzor,     # --no-pyzor
           "no-dcc" => \$no_dcc,         # --no-dcc
           "no-spamcop" => \$no_spamcop, # --no-spamcop
           "spam-only" => \$spamonly,    # --spam-only
           "ham-only" => \$hamonly,      # --ham-only
           "help" => \$help,             # --help
           "debug:s" => \$debug,         # --debug, --debug=string or --debug string
           "quiet" => \$quiet);          # --quiet

# Display usage information
if ($help) {
    output("process-quarantine.pl\n" .
           "   --ham-only      : only process confirmed ham (non-spam) items\n" .
           "   --spam-only     : only process confirmed spam items\n" .
           "   --learn         : train the Bayes database\n" .
           "   --report        : report spam to Razor/DCC/Pyzor/SpamCop\n" .
           "   --no-razor      : don't report to Razor\n" .
           "   --no-pyzor      : don't report to Pyzor\n" .
           "   --no-dcc        : don't report to DCC\n" .
           "   --no-spamcop    : don't report to SpamCop\n" .
           "   --limit=n       : process items in batches of n at a time (1-127)\n" .
           "   --max-size=n    : items larger than n (bytes) will not be processed\n" .
           "   --help          : display this help text\n" .
           "   --debug [level] : display detailed debugging information (levels: 'info' and 'all')\n" .
           "   --quiet         : display only error messages");
    exit;
}

if (!defined($debug)) {
    $debug = 0;
} elsif ($debug eq "") {
    $debug = "info";
}

# Resolve any debug/quiet conflicts
if ($debug && $quiet) {
    $debug = 0;
    $quiet = 0;
    output("Warning: --debug and --quiet negate each other.");
}

# Check for existing lock file, remove any stale locks
if (open(PID_FILE, "< $pid_file\0")) {
    my $pid;
    while (<PID_FILE>) { chomp; $pid = $_ if /^\d+\z/ }
    close(PID_FILE) or fatal(sprintf("Can't close file %s: %s", $pid_file, $!));
    if (!defined($pid)) {
        if ($debug) {
            output("Removing invalid lock file...");
        }
        unlink($pid_file);
    } elsif (kill 0 => $pid) {
        if (!$quiet) {
            output(sprintf("Another instance [%d] is currently running.", $pid));
        }
        exit;
    } else {
        if ($debug) {
            output(sprintf("Removing stale lock file [%d]...", $pid));
        }
        unlink($pid_file);
    }
}

# Write lock file, store PID
open(PID_FILE, "> $pid_file\0") or fatal(sprintf("Can't write lock file %s", $pid_file));
print PID_FILE "$$";
close(PID_FILE);

my $dbh;

# database configuration
if (defined($dsn) && defined($username) && defined($password)) {
    $dbh = DBI->connect($dsn, $username, $password)
        or fatal("Can't connect to the Maia database (verify \$dsn, \$username, and \$password in maia.conf)");
} else {
    fatal("Can't connect to the Maia database (missing \$dsn, \$username, or \$password in maia.conf)");
}

# Assemble command-line to call the subroutine
my $subroutine = $script_dir . "/process-quarantine-sub.pl";
my @cl_array = ($subroutine);
push(@cl_array, "--limit", $limit) if ($limit); # limit
push(@cl_array, "--max-size", $max_size) if ($max_size); # max-size
push(@cl_array, "--learn") if ($learn); # train the Bayes database
push(@cl_array, "--report") if ($report); # report spam
push(@cl_array, "--no-razor") if ($no_razor); # don't report to Razor
push(@cl_array, "--no-pyzor") if ($no_pyzor); # don't report to Pyzor
push(@cl_array, "--no-dcc") if ($no_dcc); # don't report to DCC
push(@cl_array, "--no-spamcop") if ($no_spamcop); # don't report to SpamCop
push(@cl_array, "--spam-only") if ($spamonly); # process spam only
push(@cl_array, "--ham-only") if ($hamonly); # process ham (non-spam) only
push(@cl_array, "--debug", $debug) if ($debug); # turn on debugging mode
push(@cl_array, "--quiet") if ($quiet); # turn on quiet mode

# Call the subroutine until there are no items left to process
# or we encounter a fatal error (i.e. return code 255)
my $total_items_processed = 0;
my $total_processing_runs = 0;
my $items_processed = -1;
while ($items_processed != 0) {
    my $clout = system(@cl_array) / 256;
    fatal("Subroutine terminated with a fatal error.") if ($clout == 255);
    $items_processed = $clout;
    $total_items_processed += $items_processed;
    $total_processing_runs++ if ($items_processed > 0);
    if ($debug) {
        output(sprintf("Subroutine processed %d %s.", $items_processed, 
              ($items_processed == 1 ? " item" : " items")));
    }
}
if ($total_items_processed > 0) {
    if ($debug) {
        output("Recalculating statistics");
    }
    recalc_stats($dbh);
}
if (!$quiet) {
    output(sprintf("Processing completed: %d %s processed in %d %s.", 
           $total_items_processed, ($total_items_processed == 1 ? "item" : "items"),
           $total_processing_runs, ($total_processing_runs == 1 ? "run" : "runs")));
}

# Disconnect from the database
$dbh->disconnect;

unlink($pid_file);

exit;


# Die, printing a time-stamped error message.
sub fatal($) {
    my ($msg) = @_;

    output("FATAL ERROR: " . $msg);
    exit 1;
}


# Write a time-stamped string to stdout for logging purposes.
sub output($) {
    my ($msg) = @_;
    my ($year, $month, $day, $hour, $minute, $second);
    my ($second, $minute, $hour, $day, $month, $year) = (localtime)[0,1,2,3,4,5];

    printf("%04d-%02d-%02d %02d:%02d:%02d Maia: [%s] %s\n",
           $year+1900, $month+1, $day, $hour, $minute, $second, $script_name, $msg);
}


# Recalculate the suspected spam OR suspected ham statistics
# for a specific user.
sub update_mail_stats($$$) {
    my($dbh, $user_id, $type) = @_;
    my($sth, $sth2, $sth3, $sth4, @row, $select, $update, $insert);
    my($token);
      
    if ($type eq "suspected_spam") {
        $token = "S";
    } elsif ($type eq "suspected_ham") {
        $token = "H";
    } else {
        $token = "";
    }

    # Calculate current aggregates for items of the specified type
    $select = "SELECT MIN(received_date) AS mindate, " .
                     "MAX(received_date) AS maxdate, " .
                     "MIN(score) AS minscore, " .
                     "MAX(score) AS maxscore, " .
                     "SUM(score) AS totalscore, " .
                     "MIN(size) AS minsize, " .
                     "MAX(size) AS maxsize, " .
                     "SUM(size) AS totalsize, " .
                     "COUNT(id) AS items " .
              "FROM maia_mail, maia_mail_recipients " .
              "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
              "AND maia_mail_recipients.type = ? " .
              "AND maia_mail_recipients.recipient_id = ? ";
    $sth = $dbh->prepare($select)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute($token, $user_id)
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    my($mindate, $maxdate, $minscore, $maxscore, $totalscore, $minsize, $maxsize, $totalsize, $items);
    if (($mindate, $maxdate, $minscore, $maxscore, $totalscore, $minsize, $maxsize, $totalsize, $items) = $sth->fetchrow()) {
        $select = "SELECT user_id FROM maia_stats WHERE user_id = ?";
        $sth2 = $dbh->prepare($select)
                    or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
        $sth2->execute($user_id)
             or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));

        # User already has a stats record, update it.
        if ($user_id == $sth2->fetchrow()) {
            $update = "UPDATE maia_stats SET oldest_" . $type . "_date = ?, " .
                                            "newest_" . $type . "_date = ?, " .
                                            "lowest_" . $type . "_score = ?, " .
                                            "highest_" . $type . "_score = ?, " .
                                            "total_" . $type . "_score = ?, " .
                                            "smallest_" . $type . "_size = ?, " .
                                            "largest_" . $type . "_size = ?, " .
                                            "total_" . $type . "_size = ?, " .
                                            "total_" . $type . "_items = ? " .
                      "WHERE user_id = ?";
                                              
           $sth3 = $dbh->prepare($update)
                       or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
           $sth3->execute($mindate,
                          $maxdate,
                          (defined($minscore) ? $minscore : 0),
                          (defined($maxscore) ? $maxscore : 0),
                          (defined($totalscore) ? $totalscore : 0),
                          (defined($minsize) ? $minsize : 0),
                          (defined($maxsize) ? $maxsize : 0),
                          (defined($totalsize) ? $totalsize : 0),
                          (defined($items) ? $items : 0),
                          $user_id)
                or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
           $sth3->finish;

        # User doesn't have a stats record yet, create a new one for him.
        } else {
            $insert = "INSERT INTO maia_stats (oldest_" . $type . "_date, " .
                                              "newest_" . $type . "_date, " .
                                              "lowest_" . $type . "_score, " .
                                              "highest_" . $type . "_score, " .
                                              "total_" . $type . "_score, " .
                                              "smallest_" . $type . "_size, " .
                                              "largest_" . $type . "_size, " .
                                              "total_" . $type . "_size, " .
                                              "total_" . $type . "_items, " .
                                              "user_id) " .
                      "VALUES (?,?,?,?,?,?,?,?,?,?)";
            $sth4 = $dbh->prepare($insert)
                        or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
            $sth4->execute($mindate,
                           $maxdate,
                           (defined($minscore) ? $minscore : 0),
                           (defined($maxscore) ? $maxscore : 0),
                           (defined($totalscore) ? $totalscore : 0),
                           (defined($minsize) ? $minsize : 0),
                           (defined($maxsize) ? $maxsize : 0),
                           (defined($totalsize) ? $totalsize : 0),
                           (defined($items) ? $items : 0),
                           $user_id)
                or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
            $sth4->finish;
        }
        $sth2->finish;
    }
    $sth->finish;
}


# Recalculate the suspected spam and suspected ham statistics
# for all users.
sub recalc_stats($) {
    my($dbh) = @_;
    my($sth, @row, $select, $user_id);

    $select = "SELECT maia_users.id FROM maia_users";
    $sth = $dbh->prepare($select)
               or fatal(sprintf("Couldn't prepare query: %s", $dbh->errstr));
    $sth->execute()
        or fatal(sprintf("Couldn't execute query: %s", $dbh->errstr));
    while (@row = $sth->fetchrow_array()) {
        $user_id = $1 if $row[0] =~ /^([0-9]+[0-9]*)$/si; # untaint
        update_mail_stats($dbh, $user_id, "suspected_spam");
        update_mail_stats($dbh, $user_id, "suspected_ham");
    }
      
    $sth->finish;
}
